/*
 *  ======== Server.c ========
 *
 */

/* this define must precede inclusion of any xdc header file */
#define Registry_CURDESC Test__Desc
#define MODULE_NAME "Server"

#include <xdc/std.h>
#include <xdc/cfg/global.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Registry.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/knl/Cache.h>

#include <ti/ipc/HeapBufMP.h>
#include <ti/ipc/MessageQ.h>
#include <ti/ipc/MultiProc.h>
#include <ti/ipc/Notify.h>
#include <ti/ipc/SharedRegion.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/knl/Task.h>

/* local header files */
#include "../cfg/AppCommon.h"
#include "../cfg/SystemCfg.h"

/* module header file */
#include "Server.h"

#define QUEUESIZE 8


/* module structure */
typedef struct {
    UInt16              hostProcId;         // host processor id
    UInt16              lineId;             // notify line id
    UInt32              eventId;            // notify event id
    Semaphore_Struct    semObj;             // semaphore object
    Semaphore_Handle    semH;               // handle to object above
    UInt32              eventQue[QUEUESIZE];  // command queue
    UInt                head;               // queue head pointer
    UInt                tail;               // queue tail pointer
    UInt32              error;
    HeapBufMP_Handle    msgHeap;            // opened remotely
    MessageQ_Handle     dspQue;             // created locally
} Server_Object;

/* private functions */
Void Server_notifyCB(UInt16 procId, UInt16 lineId, UInt32 eventId, UArg arg,
        UInt32 payload);

static UInt32   Server_waitForEvent(UInt32 wait);

/* private data */
Registry_Desc           Registry_CURDESC;
static Int              curInit = 0;
static Server_Object    Module;


/*
 *  ======== Server_init ========
 */
Void Server_init(Void)
{
    Registry_Result     result;

    if (curInit++ != 0) {
        return;  /* already initialized */
    }

    /* register with xdc.runtime to get a diags mask */
    result = Registry_addModule(&Registry_CURDESC, MODULE_NAME);
    Assert_isTrue(result == Registry_SUCCESS, (Assert_Id)NULL);

    /* initialize module object state */
    Module.hostProcId = MultiProc_getId(Global_HostProc);
    Module.lineId = SystemCfg_LineId;
    Module.eventId = SystemCfg_AppEventId;
    Module.semH = NULL;
    Module.head = 0;
    Module.tail = 0;
    Module.error = 0;
}


/*
 *  ======== Server_exit ========
 */
Void Server_exit(Void)
{
//  Registry_Result     result;

    if (curInit-- != 1) {
        return;  /* object still being used */
    }

    /* unregister from xdc.runtime */
//  result = Registry_removeModule(MODULE_NAME);
//  Assert_isTrue(result == Registry_SUCCESS, (Assert_Id)NULL);
}


/*
 *  ======== Server_setup ========
 *
 *  1. create sync object
 *  2. register notify callback
 *  3. wait until remote core has also registered notify callback
 *
 *  4. create local & shared resources (to be opened by remote processor)
 *  5. send resource ready event
 *  6. wait for remote resource ready event
 *
 *  7. open remote resources
 *  8. send application ready event
 *  9. wait for remote server ready event
 */
Int Server_setup(Void)
{
    Int                 status;
    UInt32              event;
    Semaphore_Params    semParams;
    MessageQ_Params     msgqParams;
    UInt16              procId = Module.hostProcId;
    UInt16              lineId = Module.lineId;
    UInt32              eventId = Module.eventId;

    /* enable some log events */
    Diags_setMask(MODULE_NAME"+EXF");
    Log_print0(Diags_ENTRY, "--> Server_setup:");

    /* 1. create sync object */
    Semaphore_Params_init(&semParams);
    semParams.mode = Semaphore_Mode_COUNTING;
    Semaphore_construct(&Module.semObj, 0, &semParams);
    Module.semH = Semaphore_handle(&Module.semObj);

    /* 2. register notify callback */
    status = Notify_registerEventSingle(procId, lineId,
        eventId, Server_notifyCB, (UArg)&Module);

    if (status < 0) {
        goto leave;
    }

    /* 3. wait until remote core has also registered notify callback */
    do {
        status = Notify_sendEvent(procId, lineId, eventId, App_CMD_NOP, TRUE);

        if (status == Notify_E_EVTNOTREGISTERED) {
            Task_sleep(200);
        }
    } while (status == Notify_E_EVTNOTREGISTERED);

    if (status < 0) {
        goto leave;
    }

    /* 4. create local & shared resources (to be opened by remote processor) */

    /* create local message queue (inbound messages) */
    MessageQ_Params_init(&msgqParams);

    Module.dspQue = MessageQ_create(App_DspMsgQueName, &msgqParams);

    if (Module.dspQue == NULL) {
        status = -1;
        goto leave;
    }

    /* 5. send resource ready event */
    status = Notify_sendEvent(procId, lineId, eventId, App_CMD_RESRDY, TRUE);

    if (status < 0) {
        goto leave;
    }

    /* 6. wait for remote resource ready event */
    event = Server_waitForEvent(App_CMD_RESRDY);

    if (event >= App_E_FAILURE) {
        status = -1;
        goto leave;
    }

    /* 7. open remote resources */

    /* open message heap */
    status = HeapBufMP_open(App_MsgHeapName, &Module.msgHeap);

    /* register heap with MessageQ */
    status = MessageQ_registerHeap((Ptr)(Module.msgHeap), App_MsgHeapId);

    if (status < 0) {
        goto leave;
    }

    /* 8. send application ready event */
    status = Notify_sendEvent(procId, lineId, eventId, App_CMD_READY, TRUE);

    if (status < 0) {
        goto leave;
    }

    /* 9. wait for remote server ready event */
    event = Server_waitForEvent(App_CMD_READY);

    if (event >= App_E_FAILURE) {
        status = -1;
        goto leave;
    }

leave:
    if (status < 0) {
        Log_error1("Server_setup: error=0x%x", (IArg)status);
    }
    Log_print1(Diags_EXIT, "<-- Server_setup: %d", (IArg)status);
    return(status);
}


/*
 *  ======== Server_finish ========
 *
 *  1. close remote resources
 *  2. send close done event
 *  3. wait for remote close done event
 *
 *  4. delete shared resoures
 *  5. send disconnect event (last sent event)
 *  6. wait for disconnect event
 *
 *  7. unregister notify callback
 *  8. delete sync object
 */
Int Server_finish(Void)
{
    Int         status;
    UInt32      event;
    UInt16      procId = Module.hostProcId;
    UInt16      lineId = Module.lineId;
    UInt32      eventId = Module.eventId;

    Log_print0(Diags_ENTRY, "--> Server_finish:");

    /* 1. close remote resources */

    /* unregister heap with MessageQ */
    status = MessageQ_unregisterHeap(App_MsgHeapId);

    if (status < 0) {
        goto leave;
    }

    /* close message heap */
    status = HeapBufMP_close(&Module.msgHeap);

    if (status < 0) {
        goto leave;
    }

    /* 2. send close done event */
    status = Notify_sendEvent(procId, lineId, eventId, App_CMD_CLOSED, TRUE);

    if (status < 0) {
        goto leave;
    }

    /* 3. wait for remote close done event */
    event = Server_waitForEvent(App_CMD_CLOSED);

    if (event >= App_E_FAILURE) {
        status = -1;
        goto leave;
    }

    /* 4. delete shared resoures */

    /* delete the dsp message queue */
    status = MessageQ_delete(&Module.dspQue);

    if (status < 0) {
        goto leave;
    }

    /* 5. send disconnect event (last sent event) */
    status = Notify_sendEvent(procId, lineId, eventId, App_CMD_DONE, TRUE);

    if (status < 0) {
        goto leave;
    }

    /* 6. wait for disconnect event (last event received) */
    event = Server_waitForEvent(App_CMD_DONE);

    if (event >= App_E_FAILURE) {
        status = -1;
        goto leave;
    }

    /* 7. unregister notify callback */
    status = Notify_unregisterEventSingle(procId, lineId, eventId);

    /* 8. delete sync object */
    Semaphore_destruct(&Module.semObj);
    Module.semH = NULL;

    if (status < 0) {
        goto leave;
    }

leave:
    if (status < 0) {
        Log_error1("Server_finish: error=0x%x", (IArg)status);
    }

    /* disable log events */
    Log_print1(Diags_EXIT, "<-- Server_finish: %d", (IArg)status);
    Diags_setMask(MODULE_NAME"-EXF");

    return(status);
}


/*
 *  ======== Server_run ========
 */
Int Server_run(Void)
{
    Int         status;
    Bool        running = TRUE;
//  UInt32      cmd;
    App_Msg *   msg;

    Log_print0(Diags_ENTRY, "--> Server_exec:");

    while (running) {

        /* wait for inbound message */
        status = MessageQ_get(Module.dspQue, (MessageQ_Msg *)&msg,
            MessageQ_FOREVER);

        if (status == MessageQ_E_UNBLOCKED) {
            running = FALSE;

            /* send shutdown acknowledgement */
            status = Notify_sendEvent(Module.hostProcId, Module.lineId,
                Module.eventId, App_CMD_SDACK, TRUE);
            break;
        }
        else if (status < 0) {
            goto leave;
        }

        /* process the message */
//      cmd = msg->cmd;

        /* free the message */
        status = MessageQ_free((MessageQ_Msg)msg);

        if (status < 0) {
            goto leave;
        }

        /* send data request event */
        status = Notify_sendEvent(Module.hostProcId, Module.lineId,
            Module.eventId, App_CMD_DATAREQ, TRUE);

        if (status < 0) {
            goto leave;
        }

    } /* while (running) */

leave:
    Log_print1(Diags_EXIT, "<-- Server_exec: %d", (IArg)status);
    return(status);
}


/*
 *  ======== Server_notifyCB ========
 */
Void Server_notifyCB(
    UInt16      procId,
    UInt16      lineId,
    UInt32      eventId,
    UArg        arg,
    UInt32      payload)
{
    UInt next;

    /* ignore no-op events */
    if (payload == App_CMD_NOP) {
        return;
    }

    /* handle shutdown command here */
    if ((App_CMD_MASK & payload) == App_CMD_SHUTDOWN) {
        MessageQ_unblock(Module.dspQue);
        return;
    }

    /* compute next slot in queue */
    next = (Module.head + 1) % QUEUESIZE;

    if (next == Module.tail) {
        /* queue is full, drop event and set error flag */
        Module.error = App_E_OVERFLOW;
        return;
    }
    else {
        Module.eventQue[Module.head] = payload;
        Module.head = next;
    }

    /* signal semaphore (counting) that new event is in queue */
    Semaphore_post(Module.semH);
}


/*
 *  ======== Server_waitForEvent ========
 */
static UInt32 Server_waitForEvent(UInt32 wait)
{
    UInt32 event;

    do {
        if (Module.error >= App_E_FAILURE) {
            event = Module.error;
            goto leave;
        }

        /* use counting semaphore to wait for next event */
        Semaphore_pend(Module.semH, BIOS_WAIT_FOREVER);

        /* remove next command from queue */
        event = Module.eventQue[Module.tail];
        Module.tail = (Module.tail + 1) % QUEUESIZE;

    } while ((wait != App_CMD_ANY) && ((App_CMD_MASK & event) != wait));

leave:
    return(event);
}
