/*
 *  ======== main_dsp.c ========
 *
 */

/* xdctools header files */
#include <xdc/std.h>
#include <xdc/cfg/global.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/System.h>

/* package header files */
#include <ti/ipc/Ipc.h>
#include <ti/ipc/MultiProc.h>
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

/* local header files */
#include "Server.h"

/* private functions */
static Void smain(UArg arg0, UArg arg1);


/*
 *  ======== main ========
 */
Int main(Int argc, Char* argv[])
{
    Error_Block     eb;
    Task_Params     taskParams;

    Log_print0(Diags_INFO, "--> main:");

    /* must initialize the error block before using it */
    Error_init(&eb);

    /* create main thread (interrupts not enabled in main on BIOS) */
    Task_Params_init(&taskParams);
    taskParams.instance->name = "smain";
    taskParams.arg0 = (UArg)argc;
    taskParams.arg1 = (UArg)argv;
    taskParams.stackSize = 0x1000;
    Task_create(smain, &taskParams, &eb);

    if (Error_check(&eb)) {
        System_abort("main: failed to create application startup thread");
    }

    /* start scheduler, this never returns */
    BIOS_start();
    
    /* should never get here */
    return(0);
}


/*
 *  ======== smain ========
 */
Void smain(UArg arg0, UArg arg1)
{
    Int                 status = 0;
    Error_Block         eb;
    UInt16              remoteProcId;
    Bool                running = TRUE;

    Log_print0(Diags_ENTRY | Diags_INFO, "--> smain:");

    Error_init(&eb);

    /* initialize modules */
    Server_init();

    /* turn on some trace */
//  Diags_setMask("Server+F");

    /* loop forever */
    while (running) {

        /* only one thread must call start */
        do {
            status = Ipc_start();
        } while (status == Ipc_E_NOTREADY);

        if (status < 0) {
            goto leave;
        }

        /* attach to the remote processor */
        remoteProcId = MultiProc_getId(Global_HostProc);

        /* connect to remote processor */
        do {
            status = Ipc_attach(remoteProcId);

            if (status == Ipc_E_NOTREADY) {
                Task_sleep(200);
            }
        } while (status == Ipc_E_NOTREADY);

        if (status < 0) {
            Log_error0("smain: Ipc_attach() failed");
            goto leave;
        }

        /*
         *  BEGIN server phase
         */

        /* server setup phase */
        status = Server_setup();

        if (status < 0) {
            goto leave;
        }

        /* server execute phase */
        status = Server_run();

        if (status < 0) {
            goto leave;
        }

        /* server shutdown phase */
        status = Server_finish();

        if (status < 0) {
            goto leave;
        }

        /*
         *  END server phase
         */

        /* disconnect from remote processor */
        do {
            status = Ipc_detach(remoteProcId);

            if (status == Ipc_E_NOTREADY) {
                Task_sleep(200);
            }
        } while (status == Ipc_E_NOTREADY);

        if (status < 0) {
            goto leave;
        }

        /* only one thread must call stop */
        Ipc_stop();

    } /* while (running) */


    /* finalize modules */
    Server_exit();

leave:
    Log_print1(Diags_EXIT, "<-- smain: %d", (IArg)status);
    return;
}
