/*
 *  ======== App.c ========
 *
 */

/* stdlib c header files */
#include <stdio.h>
#include <unistd.h>

/* package header files */
#include <ti/syslink/Std.h>     /* must be first */
#include <ti/ipc/HeapBufMP.h>
#include <ti/ipc/MessageQ.h>
#include <ti/ipc/MultiProc.h>
#include <ti/ipc/Notify.h>
#include <ti/ipc/SharedRegion.h>
#include <ti/syslink/utils/OsalSemaphore.h>

/* local header files */
#include "../cfg/SystemCfg.h"
#include "../cfg/AppCommon.h"
#include "App.h"

#define QUEUESIZE 8

/* module structure */
typedef struct {
    UInt16                  remoteProcId;
    UInt16                  lineId;     // notify line id
    UInt32                  eventId;    // notify event id
    OsalSemaphore_Handle    semH;
    UInt32                  eventQue[QUEUESIZE];
    UInt                    head;       // queue head pointer
    UInt                    tail;       // queue tail pointer
    UInt32                  error;
    HeapBufMP_Handle        msgHeap;    // created locally
    MessageQ_QueueId        dspQue;     // opened remotely
    UInt16                  heapId;     // MessageQ heapId
    UInt32                  msgSize;
} App_Module;

/* private functions */
static Void     App_notifyCB(
        UInt16  procId,
        UInt16  lineId,
        UInt32  eventId,
        UArg    arg,
        UInt32  payload);

static UInt32   App_waitForEvent(UInt32 wait);

/* Module state object */
static App_Module Module = {
    .remoteProcId = MultiProc_INVALIDID,
    .lineId = SystemCfg_LineId,
    .eventId = SystemCfg_AppEventId,
    .semH = NULL,
    .head = 0,
    .tail = 0,
    .error = 0,
    .msgHeap = NULL,
    .dspQue = MessageQ_INVALIDMESSAGEQ,
    .heapId = App_MsgHeapId,
    .msgSize = sizeof(App_Msg)
};


/*
 *  ======== App_start ========
 *
 *  1. create sync object
 *  2. register notify callback
 *  3. wait until remote core has also registered notify callback
 *
 *  4. create local & shared resources (to be opened by remote processor)
 *  5. send resource ready event
 *  6. wait for remote resource ready event
 *
 *  7. open remote resources
 *  8. send application ready event
 *  9. wait for remote server ready event
 */
Int App_start(Char *remoteProc)
{
    Int                 status;
    UInt32              event;
    HeapBufMP_Params    heapParams;
    UInt16              procId;
    UInt16              lineId = Module.lineId;
    UInt32              eventId = Module.eventId;


    Module.remoteProcId = MultiProc_getId(remoteProc); /* App_init? */
    procId = Module.remoteProcId;

    /* 1. create sync object */
    Module.semH = OsalSemaphore_create(OsalSemaphore_Type_Counting);

    if (Module.semH == NULL) {
        status = -1;
        goto leave;
    }

    /* 2. register notify callback */
    status = Notify_registerEventSingle(procId, lineId,
        eventId, App_notifyCB, (UArg)&Module);

    if (status < 0) {
        goto leave;
    }

    /* 3. wait until remote core has also registered notify callback */
    do {
        status = Notify_sendEvent(procId, lineId, eventId, App_CMD_NOP, TRUE);

        if (status == Notify_E_EVTNOTREGISTERED) {
            sleep(1);
        }
    } while (status == Notify_E_EVTNOTREGISTERED);

    if (status < 0) {
        goto leave;
    }

    /* 4. create local & shared resources (to be opened by remote processor) */

    /* create heap for messages */
    HeapBufMP_Params_init(&heapParams);

    heapParams.name = App_MsgHeapName;
//  heapParams.regionId = App_MsgHeapSrId;
    heapParams.regionId = 0;
    heapParams.blockSize = 64;
    heapParams.numBlocks = 10;

    Module.msgHeap = HeapBufMP_create(&heapParams);

    if (Module.msgHeap == NULL) {
        status = -1;
        goto leave;
    }

    /* register heap with MessageQ */
    status = MessageQ_registerHeap((Ptr)(Module.msgHeap), App_MsgHeapId);

    if (status < 0) {
        goto leave;
    }

    /* 5. send resource ready event */
    status = Notify_sendEvent(procId, lineId, eventId, App_CMD_RESRDY, TRUE);

    if (status < 0) {
        goto leave;
    }

    /* 6. wait for remote resource ready event */
    event = App_waitForEvent(App_CMD_RESRDY);

    if (event >= App_E_FAILURE) {
        status = -1;
        goto leave;
    }

    /* 7. open remote resources */

    /* open the dsp message queue */
    status = MessageQ_open(App_DspMsgQueName, &Module.dspQue);

    if (status < 0) {
        goto leave;
    }

    /* 8. send application ready event */
    status = Notify_sendEvent(procId, lineId, eventId, App_CMD_READY, TRUE);

    if (status < 0) {
        goto leave;
    }

    /* 9. wait for remote server ready event */
    event = App_waitForEvent(App_CMD_READY);

    if (event >= App_E_FAILURE) {
        status = -1;
        goto leave;
    }

leave:
    return(status);
}


/*
 *  ======== App_stop ========
 *
 *  1. send shutdown event
 *  2. wait for shutdown acknowledgement event
 *
 *  3. close remote resources
 *  4. send close done event
 *  5. wait for remote close done event
 *
 *  6. delete shared resoures
 *  7. send disconnect event (last sent event)
 *  8. wait for disconnect event
 *
 *  9. unregister notify callback
 * 10. delete sync object
 */
Int App_stop(Void)
{
    Int         status;
    UInt32      event;
    UInt16      procId = Module.remoteProcId;
    UInt16      lineId = Module.lineId;
    UInt32      eventId = Module.eventId;

    /* 1. send shutdown command */
    status = Notify_sendEvent(procId, lineId, eventId, App_CMD_SHUTDOWN, TRUE);

    if (status < 0) {
        goto leave;
    }

    /* 2. wait for shutdown acknowledgement */
    event = App_waitForEvent(App_CMD_SDACK);

    if (event >= App_E_FAILURE) {
        status = -1;
        goto leave;
    }

    /* 3. close remote resources */
    status = MessageQ_close(&Module.dspQue);

    if (status < 0) {
        goto leave;
    }

    /* 4. send close done event */
    status = Notify_sendEvent(procId, lineId, eventId, App_CMD_CLOSED, TRUE);

    if (status < 0) {
        goto leave;
    }

    /* 5. wait for remote close done event */
    event = App_waitForEvent(App_CMD_CLOSED);

    if (event >= App_E_FAILURE) {
        status = -1;
        goto leave;
    }

    /* 6. delete shared resoures */

    /* unregister heap with MessageQ */
    status = MessageQ_unregisterHeap(App_MsgHeapId);

    if (status < 0) {
        goto leave;
    }

    /* delete the message heap */
    status = HeapBufMP_delete(&Module.msgHeap);

    if (status < 0) {
        goto leave;
    }

    /* 7. send disconnect event (last sent event) */
    status = Notify_sendEvent(procId, lineId, eventId, App_CMD_DONE, TRUE);

    if (status < 0) {
        goto leave;
    }

    /* 8. wait for disconnect event (last event received) */
    event = App_waitForEvent(App_CMD_DONE);

    if (event >= App_E_FAILURE) {
        status = -1;
        goto leave;
    }

    /* 9. unregister notify callback */
    status = Notify_unregisterEventSingle(procId, lineId, eventId);

    if (status < 0) {
        goto leave;
    }

    /* 10. delete sync object */
    OsalSemaphore_delete(&Module.semH);

leave:
    return(status);
}


/*
 *  ======== App_exec ========
 */
Int App_exec(Void)
{
    Int         status;
    Int         i;
    App_Msg *   msg;
    UInt32      event;

    printf("--> App_exec:\n");

    /* fill process pipeline */
    for (i = 1; i <= 3; i++) {
        printf("App_exec: sending message %d\n", i);

        /* allocate message */
        msg = (App_Msg *)MessageQ_alloc(Module.heapId, Module.msgSize);

        if (msg == NULL) {
            status = -1;
            goto leave;
        }

        /* fill in message payload */
        msg->cmd = App_CMD_NOP;

        /* send message */
        MessageQ_put(Module.dspQue, (MessageQ_Msg)msg);
    }

    /* process steady state (keep pipeline full) */
    for (i = 4; i <= 15; i++) {

        /* wait for data request */
        event = App_waitForEvent(App_CMD_DATAREQ);

        if (event >= App_E_FAILURE) {
            status = -1;
            goto leave;
        }

        printf("App_exec: sending message %d\n", i);

        /* allocate message */
        msg = (App_Msg *)MessageQ_alloc(Module.heapId, Module.msgSize);

        if (msg == NULL) {
            status = -1;
            goto leave;
        }

        /* fill in message payload */
        msg->cmd = App_CMD_NOP;

        /* send message */
        MessageQ_put(Module.dspQue, (MessageQ_Msg)msg);
    }

leave:
    printf("<-- App_exec: %d\n", status);
    return(status);
}


/*
 *  ======== App_notifyCB ========
 */
static Void App_notifyCB(
    UInt16      procId,
    UInt16      lineId,
    UInt32      eventId,
    UArg        arg,
    UInt32      payload)
{
    UInt next;

    /* ignore no-op events */
    if (payload == App_CMD_NOP) {
        return;
    }

    /* compute next slot in queue */
    next = (Module.head + 1) % QUEUESIZE;

    if (next == Module.tail) {
        /* queue is full, drop event and set error flag */
        Module.error = App_E_OVERFLOW;
        return;
    }
    else {
        Module.eventQue[Module.head] = payload;
        Module.head = next;
    }

    /* signal semaphore (counting) that new event is in queue */
    OsalSemaphore_post(Module.semH);
}


/*
 *  ======== App_waitForEvent ========
 */
static UInt32 App_waitForEvent(UInt32 wait)
{
    UInt32 event;

    do {
        if (Module.error >= App_E_FAILURE) {
            event = Module.error;
            goto leave;
        }

        /* use counting semaphore to wait for next event */
        OsalSemaphore_pend(Module.semH, OSALSEMAPHORE_WAIT_FOREVER);

        /* remove next command from queue */
        event = Module.eventQue[Module.tail];
        Module.tail = (Module.tail + 1) % QUEUESIZE;

    } while ((wait != App_CMD_ANY) && ((App_CMD_MASK & event) != wait));

leave:
    return(event);
}
