/*
 *  ======== main_host.c ========
 *
 */

/* cstdlib header files */
#include <stdio.h>
#include <stdlib.h>

/* package header files */
#include <ti/syslink/Std.h>     /* must be first */
#include <ti/syslink/IpcHost.h>
#include <ti/syslink/ProcMgr.h>
#include <ti/syslink/SysLink.h>

#include <ti/ipc/MultiProc.h>

/* local header files */
#include "App.h"

/* private functions */
static Int Main_main(Void);
static Int Main_parseArgs(Int argc, Char *argv[]);
static Int Main_procStart(String proc, String exec);
static Int Main_procStop(String proc, String exec);


#define Main_USAGE "\
Usage:\n\
    test [options] proc [executable]\n\
\n\
Arguments:\n\
    proc      : the name of the remote processor\n\
    executable: load and run the executable on processor\n\
\n\
Options:\n\
    h           : print this help message\n\
    l           : list the available remote names\n\
    p           : pause at interesting points\n\
    w           : wait before shutdown\n\
\n\
Examples:\n\
    test DSP\n\
    test DSP test_dsp.x\n\
    test -l\n\
    test -h\n\
\n"

/* private data */
static String   Main_remoteProc = NULL;
static String   Main_execProc1 = NULL;
static Bool     Main_pause = FALSE;
static Int      Main_wait = FALSE;

/*
 *  ======== main ========
 */
Int main(Int argc, Char* argv[])
{
    Int status;

    /* parse command line */
    status = Main_parseArgs(argc, argv);

    if (status < 0) {
        goto leave;
    }

    if (Main_pause) {
        printf("After Main_parseArgs. Press enter to continue...");
        getchar();
    }

    /* SysLink initialization */
    SysLink_setup();

    /* application setup, execute, shutdown */
    status = Main_main();

    if (status < 0) {
        goto leave;
    }

    /* SysLink finalization */
    SysLink_destroy();

leave:
    status = (status >= 0 ? 0 : status);
    return(status);
}


/*
 *  ======== Main_main ========
 */
Int Main_main(Void)
{
    UInt16      remoteProcId;
    Int         status = 0;


    /* load and run remote processor if executable given */
    if (Main_execProc1 != NULL) {
        status = Main_procStart(Main_remoteProc, Main_execProc1);

        if (status < 0) {
            goto leave;
        }

        if (Main_pause) {
            printf("After slave startup. Press enter to continue...");
            getchar();
        }
    }

    /* invoke the SysLink load callback */
    remoteProcId = MultiProc_getId(Main_remoteProc);
    status = Ipc_control(remoteProcId, Ipc_CONTROLCMD_LOADCALLBACK, NULL);

    if (status < 0) {
        goto leave;
    }

    /* invoke the SysLink start callback */
    status = Ipc_control(remoteProcId, Ipc_CONTROLCMD_STARTCALLBACK, NULL);

    if (status < 0) {
        goto leave;
    }

    if (Main_pause) {
        printf("After start callback. Press enter to continue...");
        getchar();
    }

    /*
     *  BEGIN application phase
     */

    /* application startup phase */
    App_start(Main_remoteProc);

    /* application execute phase */
    App_exec();

    /* application stop phase */
    App_stop();

    /*
     *  END application phase
     */

    if (Main_pause) {
        printf("After application phase. Press enter to continue...");
        getchar();
    }

    /* invoke the SysLink stop callback */
    status = Ipc_control(remoteProcId, Ipc_CONTROLCMD_STOPCALLBACK, NULL);

    if (status < 0) {
        goto leave;
    }

    if (Main_pause) {
        printf("After stop callback. Press enter to continue...");
        getchar();
    }

    /* if executable given, stop processor */
    if (Main_execProc1 != NULL) {
        status = Main_procStop(Main_remoteProc, Main_execProc1);

        if (status < 0) {
            goto leave;
        }

        if (Main_pause) {
            printf("After slave stop. Press enter to continue...");
            getchar();
        }
    }

leave:
    status = (status >= 0 ? 0 : status);
    return(status);
}


/*
 *  ======== Main_parseArgs ========
 */
Int Main_parseArgs(Int argc, Char *argv[])
{
    Int             x, cp, opt, argNum;
    UInt16          i, numProcs;
    String          name;
    Int             status = 0;


    /* parse the command line options */
    for (opt = 1; (opt < argc) && (argv[opt][0] == '-'); opt++) {
        for (x = 0, cp = 1; argv[opt][cp] != '\0'; cp++) {
            x = (x << 8) | (int)argv[opt][cp];
        }

        switch (x) {
            case 'h': /* -h */
                printf("%s", Main_USAGE);
                exit(0);
                break;

            case 'l': /* -l */
                printf("Processor List\n");
                SysLink_setup();
                numProcs = MultiProc_getNumProcessors();
                for (i = 0; i < numProcs; i++) {
                    name = MultiProc_getName(i);
                    printf("    procId=%d, procName=%s\n", i, name);
                }
                SysLink_destroy();
                exit(0);
                break;

            case 'p': /* -p */
                Main_pause = TRUE;
                break;

            case 'w': /* -w */
                Main_wait = TRUE;
                break;

            default:
                printf(
                    "Error: %s, line %d: invalid option, %c\n",
                    __FILE__, __LINE__, (Char)x);
                printf("%s", Main_USAGE);
                status = -1;
                goto leave;
        }
    }

    /* parse the command line arguments */
    for (argNum = 1; opt < argc; argNum++, opt++) {

        switch (argNum) {
            case 1: /* name of proc #1 */
                Main_remoteProc = argv[opt];
                break;

            case 2: /* executable for proc #1 */
                Main_execProc1 = argv[opt];
                break;

            default:
                printf(
                    "Error: %s, line %d: too many arguments\n",
                    __FILE__, __LINE__);
                printf("%s", Main_USAGE);
                status = -1;
                goto leave;
        }
    }

    /* validate command line arguments */
    if (Main_remoteProc == NULL) {
        printf(
            "Error: %s, line %d: missing argument: proc\n",
            __FILE__, __LINE__);
        printf("%s", Main_USAGE);
        status = -1;
        goto leave;
    }

leave:
    return(status);
}


/*
 *  ======== Main_procStart ========
 */
Int Main_procStart(String proc, String file)
{
    Int                     status = 0;
    ProcMgr_Handle          handle = NULL;
    ProcMgr_AttachParams    attachParams;
    ProcMgr_StartParams     startParams;
    String                  args[1];
    UInt16                  procId;
    UInt32                  fileId;


    printf("--> Main_procStart:\n");

    procId = MultiProc_getId(proc);
    status = ProcMgr_open(&handle, procId);

    if (status < 0) {
        printf("ProcMgr_open failed\n");
        goto leave;
    }

    ProcMgr_getAttachParams(NULL, &attachParams);
    status = ProcMgr_attach(handle, &attachParams);

    if (status < 0) {
        printf("ProcMgr_attach failed\n");
        goto leave;
    }

    args[0] = file;
    status = ProcMgr_load(handle, file, 1, args, NULL, &fileId);

    if (status < 0) {
        printf("Error in ProcMgr_load [0x%x]\n", status);
        goto leave;
    }

    ProcMgr_getStartParams(handle, &startParams);
    status = ProcMgr_start(handle, &startParams);

    if (status < 0) {
        printf("ProcMgr_start failed [0x%x]\n", status);
        goto leave;
    }

    status = ProcMgr_close(&handle);

    if (status < 0) {
        printf("ProcMgr_close() failed\n");
        goto leave;
    }

leave:
    printf("<-- Main_procStart: 0x%x\n", status);
    return(status);
}


/*
 *  ======== Main_procStop ========
 */
Int Main_procStop(String proc, String file)
{
    Int                     status = 0;
    ProcMgr_Handle          handle = NULL;
    UInt16                  procId;


    printf("--> Main_procStop:\n");

    procId = MultiProc_getId(proc);
    status = ProcMgr_open(&handle, procId);

    if (status < 0) {
        printf("ProcMgr_open failed\n");
        goto leave;
    }

    status = ProcMgr_stop(handle);

    if (status < 0) {
        printf("ProcMgr_stop failed [0x%x]\n", status);
        goto leave;
    }

    status = ProcMgr_detach(handle);

    if (status < 0) {
        printf("ProcMgr_detach failed\n");
        goto leave;
    }

    status = ProcMgr_close(&handle);

    if (status < 0) {
        printf("ProcMgr_close() failed\n");
        goto leave;
    }

leave:
    printf("<-- Main_procStop: 0x%x\n", status);
    return(status);
}
